@@ -1,4 +1,4 @@
-/* $Id: pl.c,v 1.21 2005/06/30 01:22:33 jwp Exp $
+/* $Id: pl.c,v 1.25 2005/08/10 19:06:50 jwp Exp $
  *
  * Copyright 2005, PostgresPy Project.
  * http://python.projects.postgresql.org
@@ -17,7 +17,6 @@
 #include <compile.h>
 #include <marshal.h>
 #include <structmember.h>
-#include <cStringIO.h>
 
 #include <postgres.h>
 #include <fmgr.h>
@@ -251,11 +250,78 @@ linecache_update(PyObj self, PyObj args)
 
 	return(rob);
 }
-static PyMethodDef linecache_open_def = {
-	"pg_linecache_update",
-	(PyCFunction) linecache_update,
+static PyMethodDef linecache_update_def = {
+	"pg_linecache_update", (PyCFunction) linecache_update, METH_VARARGS,
+	"custom updater allows linecache to see PostgreSQL pg_proc files"
+};
+static PyObj linecache_update_obj;
+static int
+initialize_linecache(void)
+{
+	PyObj lc_mod = PyImport_ImportModule("linecache");
+	if (lc_mod == NULL) return(-1);
+
+	if (lc_updatecache == NULL)
+	{
+		lc_updatecache = PyObject_GetAttrString(lc_mod, "updatecache");
+		if (lc_updatecache == NULL) goto fail;
+	}
+
+	if (lc_cache == NULL)
+	{
+		lc_cache = PyObject_GetAttrString(lc_mod, "cache");
+		if (lc_cache == NULL) goto fail;
+	}
+
+	if (linecache_update_obj == NULL)
+	{
+		linecache_update_obj = PyCFunction_New(&linecache_update_def, lc_mod);
+		if (linecache_update_obj == NULL) goto fail;
+	}
+
+	if (PyObject_SetAttrString(lc_mod, "updatecache", linecache_update_obj) < 0)
+		goto fail;
+
+	Py_DECREF(lc_mod);
+	return(0);
+fail:
+	Py_DECREF(lc_mod);
+	return(-1);
+}
+
+static PyObj
+warnings_showwarning(PyObj self, PyObj args, PyObj kw)
+{
+	static char *kwords[] = {
+		"message", "category", "filename", "lineno", "file", NULL
+	};
+	PyObj msg, cat, file, obstr;
+	char *filename;
+	long lineno;
+
+	if (PyArg_ParseTupleAndKeywords(args, kw, "OOsd|O", kwords,
+			&msg, &cat, &filename, &lineno, &file) < 0)
+		return(NULL);
+
+	obstr = PyObject_Str(msg);
+	if (obstr != NULL)
+	{
+		errstart(WARNING, filename, lineno, PG_FUNCNAME_MACRO);
+		errmsg(PyString_AS_STRING(obstr));
+		errfinish(0);
+		Py_DECREF(obstr);
+	}
+	else
+		return(NULL);
+
+	Py_INCREF(Py_None);
+	return(Py_None);
+}
+static PyMethodDef warnings_showwarning_def = {
+	"pg_warnings_showwarning",
+	(PyCFunction) warnings_showwarning,
 	METH_VARARGS|METH_KEYWORDS,
-	"custom updater allow linecache to see PostgreSQL pg_proc files"
+	"custom showwarning displays the warning using ereport"
 };
 
 /*
@@ -265,40 +331,68 @@ static char *
 tbtostr(PyObj tb)
 {
 	static PyObj tb_print_tb = NULL;
+	static PyObj StringIO = NULL;
 
-	PyObj rob;
-	/*
-	 * WARNING: Must be kept in sync with python:Modules/cStringIO.c
-	 */
-	struct {
-		PyObject_HEAD
-		char *buf;
-		int pos, string_size;
-		int buf_size, softspace;
-	} io = {0,};
+	PyObj io, ob;
+	char *rs = NULL;
 
 	Assert(tb != NULL && PyTraceBack_Check(tb));
 
 	if (tb_print_tb == NULL)
 	{
 		PyObj mod_traceback = PyImport_ImportModule("traceback");
-		if (mod_traceback == NULL) return(NULL);
+		if (mod_traceback == NULL) goto fail;
 
 		tb_print_tb = PyObject_GetAttrString(mod_traceback, "print_tb");
 		Py_DECREF(mod_traceback);
-		if (!tb_print_tb) return(NULL);
+		if (!tb_print_tb) goto fail;
 	}
 
-	io.ob_refcnt = 1;
-	io.ob_type = PycStringIO->OutputType;
-	io.buf = malloc(128);
-	io.buf_size = 128;
-	rob = Py_Call(tb_print_tb, tb, Py_None, &io);
-	Py_XDECREF(rob);
-	PycStringIO->cwrite((PyObj) &io, "\0", 1);
+	if (StringIO == NULL)
+	{
+		PyObj mod_stringio = PyImport_ImportModule("StringIO");
+		if (mod_stringio == NULL) goto fail;
 
+		StringIO = PyObject_GetAttrString(mod_stringio, "StringIO");
+		Py_DECREF(mod_stringio);
+		if (!StringIO) goto fail;
+	}
+	
+	io = Py_Call(StringIO);
+	if (io == NULL) goto fail;
+
+	ob = Py_Call(tb_print_tb, tb, Py_None, io);
+	if (ob == NULL)
+		goto DECREF_io;
+	else
+		Py_DECREF(ob);
+
+	ob = PyObject_CallMethod(io, "seek", "i", 0);
+	if (ob == NULL)
+		goto DECREF_io;
+	else
+		Py_DECREF(ob);
+
+	ob = PyObject_CallMethod(io, "read", "");
+	if (ob == NULL)
+		goto DECREF_io;
+
+	if (!PyString_Check(ob))
+	{
+		Py_DECREF(ob);
+		goto DECREF_io;
+	}
+
+	rs = strdup(PyString_AS_STRING(ob));
+	Py_DECREF(ob);
+	Py_DECREF(io);
+
+	return(rs);
+DECREF_io:
+	Py_DECREF(io);
+fail:
 	if (PyErr_Occurred()) PyErr_Clear();
-	return(io.buf);
+	return(NULL);
 }
 
 /*
@@ -361,6 +455,15 @@ unsigned int ist_count;
 static PyObj
 ist_begin(PyObj self)
 {
+	static bool warned = false;
+	if (warned == false)
+	{
+		warned = true;
+		if (PyErr_Warn(PyExc_DeprecationWarning,
+			"BEGIN, RELEASE, and ROLLBACK are scheduled for removal") < 0)
+			return(NULL);
+	}
+
 	PG_TRY();
 	{
 		if (ist_count == -1)
@@ -457,8 +560,9 @@ PyObj XactDict;
 static void
 XactHook(XactEvent xev, void *arg)
 {
-	while (PyList_GET_SIZE(fnExtraCalls) > 0)
-		PySequence_DelItem(fnExtraCalls, 0);
+	if (fnExtraCalls != NULL)
+		while (PyList_GET_SIZE(fnExtraCalls) > 0)
+			PySequence_DelItem(fnExtraCalls, 0);
 	ist_count = 0;
 
 	PyDict_Clear(XactDict);
@@ -498,109 +602,49 @@ CodeFixer(PyObj self)
  * Allowing ignorance of the Python initialization state within the initializor
  * and the validator, the only entry points.
  */
-static PyObj Module = NULL;
 #ifdef __APPLE__
 __attribute__((constructor))
 #endif
 void
 _init(void)
 {
-	MemoryContext former;
 	PyObj ob;
 
-	START_CRIT_SECTION();
-
 	Py_SetProgramName("Postgres");
 	Py_Initialize();
 
-	PycString_IMPORT;
-	XactDict = PyDict_New();
-
-	former = MemoryContextSwitchTo(TopMemoryContext);
-	Module = PyImport_ImportModule(DEFAULT_POSTGRES_MODULE);
-	if (Module == NULL)
-		ereport(ERROR,(
-			errmsg("failed to import Postgres interface module \""
-				DEFAULT_POSTGRES_MODULE"\""),
-			PythonExceptionDetail
-		));
-	MemoryContextSwitchTo(former);
-
-	ob = PyObject_GetAttrString(Module, PyPgCI_CObjectName);
-	if (ob == NULL)
-		ereport(ERROR,(
-			errmsg("failed to get C Interface from extension module"),
-			PythonExceptionDetail
-		));
-	PyPgCI_Initialize(PyCObject_AsVoidPtr(ob));
-	Py_DECREF(ob);
-	PyPgCI.PL.Fixer = CodeFixer;
-
-	if (PyUnicode_SetDefaultEncoding(
-				PyEncoding_FromPgEncoding(GetDatabaseEncoding())) < 0)
-		ereport(ERROR,(
-			errmsg("failed to set default encoding"),
-			PythonExceptionDetail
-		));
-
-	ob = PyImport_ImportModule("linecache");
+	ob = PyImport_ImportModule("warnings");
 	if (ob != NULL)
 	{
-		PyObj lc_open;
-		lc_updatecache = PyObject_GetAttrString(ob, "updatecache");
-		lc_cache = PyObject_GetAttrString(ob, "cache");
-		lc_open = PyCFunction_New(&linecache_open_def, ob);
-		PyObject_SetAttrString(ob, "updatecache", lc_open);
+		PyObj wd, sw;
+		wd = PyModule_GetDict(ob);
 		Py_DECREF(ob);
-		Py_DECREF(lc_open);
-	}
-	else
-		PyErr_Clear();
+		if (wd == NULL) goto showwarning_failure;
 
-	/*
-	 * The Transaction and Savepoint types are a direct interface to PostgreSQL
-	 * transaction management, therefore it is not appropriate in the context of
-	 * a function call.
-	 */
-	PyObject_DelAttrString(Module, "Transaction");
-	PyObject_DelAttrString(Module, "Savepoint");
+		sw = PyCFunction_New(&warnings_showwarning_def, ob);
+		if (sw == NULL) goto showwarning_failure;
 
-	ob = PyImport_AddModule("__main__");
-	if (ob == NULL)
-	{
-		ereport(ERROR, (
-			errmsg("failed to get __main__ dictionary")
-		));
+		if (PyDict_SetItemString(wd, "showwarning", sw) < 0)
+		{
+			Py_DECREF(sw);
+			goto showwarning_failure;
+		}
 	}
 	else
 	{
-		PyObj item;
-		ob = PyModule_GetDict(ob);
-		PyDict_SetItemString(ob, "Postgres", Module);
-
-		item = PyCFunction_New(&ist_begin_def, NULL);
-		PyDict_SetItemString(ob, "BEGIN", item);
-
-		item = PyCFunction_New(&ist_release_def, NULL);
-		PyDict_SetItemString(ob, "RELEASE", item);
-
-		item = PyCFunction_New(&ist_rollback_def, NULL);
-		PyDict_SetItemString(ob, "ROLLBACK", item);
-
-		Py_INCREF(XactDict);
-		PyDict_SetItemString(ob, "XD", XactDict);
-	}
-
-	fnExtraCalls = PyList_New(0);
-	if (fnExtraCalls == NULL)
-	{
-		ereport(ERROR, (
-			errmsg("failed to create fnExtraCalls list")
+showwarning_failure:
+		ereport(INFO, (
+			errmsg("failed to overload warnings.showwarning"),
+			errhint(
+				"Make sure Python's sys.path has the directory containing the "
+				"standard warnings module"
+			),
+			PythonExceptionDetail
 		));
 	}
+
+	XactDict = PyDict_New();
 	RegisterXactCallback(XactHook, NULL);
-
-	END_CRIT_SECTION();
 }
 #if defined(_WIN32)
 BOOL WINAPI
@@ -1036,8 +1080,6 @@ call_function(PG_FUNCTION_ARGS)
 	return(rd);
 }
 
-static PGFunction pl;
-
 static Datum
 pl_executor(PG_FUNCTION_ARGS)
 {
@@ -1051,26 +1093,143 @@ pl_executor(PG_FUNCTION_ARGS)
 	return(rd);
 }
 
+static PyObj InterfaceModule = NULL;
+
+static void *
+initialize_main(void)
+{
+	PyObj ob;
+
+	ob = PyImport_AddModule("__main__");
+	if (ob == NULL)
+	{
+		ereport(ERROR, (
+			errmsg("failed to get __main__ dictionary")
+		));
+	}
+	else
+	{
+		PyObj item;
+		ob = PyModule_GetDict(ob);
+		PyDict_SetItemString(ob, "Postgres", InterfaceModule);
+
+		item = PyCFunction_New(&ist_begin_def, NULL);
+		PyDict_SetItemString(ob, "BEGIN", item);
+
+		item = PyCFunction_New(&ist_release_def, NULL);
+		PyDict_SetItemString(ob, "RELEASE", item);
+
+		item = PyCFunction_New(&ist_rollback_def, NULL);
+		PyDict_SetItemString(ob, "ROLLBACK", item);
+
+		Py_INCREF(XactDict);
+		PyDict_SetItemString(ob, "XD", XactDict);
+	}
+
+	return(NULL);
+}
+
+static void *
+initialize_unicode(void)
+{
+	if (PyUnicode_SetDefaultEncoding(
+				PyEncoding_FromPgEncoding(GetDatabaseEncoding())) < 0)
+		ereport(ERROR,(
+			errmsg("failed to set the default PyUnicode encoding"),
+			PythonExceptionDetail
+		));
+
+	return(initialize_main);
+}
+
+static void *
+initialize_interface(void)
+{
+	PyObj ob;
+
+	if (InterfaceModule == NULL)
+	{
+		/*
+		 * XXX: Use GUC Variable Here. (?)
+		 */
+		InterfaceModule = PyImport_ImportModule(DEFAULT_POSTGRES_MODULE);
+		if (InterfaceModule == NULL)
+			ereport(ERROR,(
+				errmsg("failed to import Postgres interface module"),
+				PythonExceptionDetail
+			));
+	}
+
+	/*
+	 * The Transaction and Savepoint types are a direct interface to PostgreSQL
+	 * transaction management, therefore it is not appropriate in the context of
+	 * a function call.
+	 *
+	 * XXX: Deprecating.
+	 */
+	PyObject_DelAttrString(InterfaceModule, "Transaction");
+	PyObject_DelAttrString(InterfaceModule, "Savepoint");
+
+	ob = PyObject_GetAttrString(InterfaceModule, PyPgCI_CObjectName);
+	if (ob == NULL)
+		ereport(ERROR,(
+			errmsg("failed to get C Interface from extension module"),
+			PythonExceptionDetail
+		));
+	PyPgCI_Initialize(PyCObject_AsVoidPtr(ob));
+	Py_DECREF(ob);
+	PyPgCI.PL.Fixer = CodeFixer;
+
+	return(initialize_unicode);
+}
+
+static void *
+initialize_python(void)
+{
+	if (!Py_IsInitialized())
+	{
+		_init();
+		if (!Py_IsInitialized())
+			ereport(ERROR,(
+				errmsg("failed to initialize Python")
+			));
+	}
+
+	return(initialize_interface);
+}
+
+Datum pl_handler(PG_FUNCTION_ARGS);
+static PGFunction pl;
 static Datum
 pl_initializor(PG_FUNCTION_ARGS)
 {
+	static void * (*init_state)(void) = initialize_python;
 	PyObj code, func;
 
-	/*
-	 * This first lookup will not have the code initialized, as
-	 * the function type does not know what is a Python language function
-	 * and what is not.
-	 */
-	PyPgCI.PL.Handler = fcinfo->flinfo->fn_addr;
-	func = PyPgFunction_FromOid(fcinfo->flinfo->fn_oid);
-	code = PyPgFunction_FetchCode(func);
-	if (CodeFixer(func) < 0)
-		ereport(ERROR,(
-			errmsg("failed to initialize procedural language")
+	while (init_state != NULL)
+		init_state = init_state();
+
+	if (initialize_linecache() < 0)
+		ereport(INFO, (
+			errmsg(
+				"failed to overload \"linecache.updatecache\" "
+				"for procedure tracebacks"
+			),
+			PythonExceptionDetail
 		));
-	Py_DECREF(code);
-	PyPgCI.PL.Oid = PyPgFunction_FetchProcStruct(func)->prolang;
-	Py_DECREF(func);
+
+	if (fnExtraCalls == NULL)
+	{
+		fnExtraCalls = PyList_New(0);
+		if (fnExtraCalls == NULL)
+		{
+			ereport(ERROR, (
+				errmsg("failed to create fnExtraCalls list")
+			));
+		}
+	}
+
+	PyPgCI.PL.Handler = pl_handler;
 
 	/*
 	 * Everything is initialized, and will be for the duration of this process.
@@ -1096,8 +1255,6 @@ Datum pl_handler(PG_FUNCTION_ARGS)
 	MemoryContext fn_mcxt = fcinfo->flinfo->fn_mcxt;
 	MemoryContext former = MemoryContextSwitchTo(TopMemoryContext);
 
-	Assert(Py_IsInitialized());
-
 	rd = pl(fcinfo);
 	fcinfo->flinfo->fn_mcxt = fn_mcxt;
 	MemoryContextSwitchTo(former);
